/************************************************************************/
/*    FCHGALL 	Program to find and replace character strings in all	*/
/*		modules in a directory.					*/
/*									*/
/* 		Author: W. A. Smethurst - 73657.2447@compuserve.com	*/
/*		Date:	07/09/95					*/
/************************************************************************/

#include "stdio.h"
#include "stdlib.h"
#include "process.h"
#include "dos.h"
#include "conio.h"
#include "ctype.h"
#include "alloc.h"
#include "dir.h"
#include "errno.h"

/************************************************************************/
/* Global Definitions							*/
/************************************************************************/

#define SUCCESS		0			  /* Function succeeded	*/
#define FAIL		-1		       /* Function unsuccessful	*/
#define EXIT		1			 /* Exit Program switch	*/
#define BKSLASH		92				    /* Hex 0x5c	*/
#define FALSE		0
#define TRUE		!FALSE
#define QUOTE		34				/* Double Quote	*/
#define RECSIZE		512

#ifndef   BOOLTYPE
  #define BOOLTYPE 0
  typedef unsigned short int BOOL;
#endif

typedef struct dfile		      /* File name structure definition	*/
	{
	char filename[13];
	struct dfile *next;		   /* Pointer to next structure	*/
	} DFILE;

/************************************************************************/
/* Global Variable Declarations						*/
/************************************************************************/

char *ProgTitle = "FCHGALL v1.0 07/09/95 (c) W. A. Smethurst";

int row,col;				/* Cursor Row and Column values	*/
int FileCnt;				   /* Number of files processed	*/
int ChgCnt;				     /* Number of files changed	*/
int ByteCnt;				/* Bytes of InBuffer to process	*/

BOOL ChangesMade;      /* Set to TRUE if any changes are made to a file	*/

int OldLength,NewLength;       /* Change string lengths (after editing)	*/

char OldString[MAXPATH],NewString[MAXPATH];	      /* Change strings	*/

int InBytes,InPos;		    /* Positioning indices for InBuffer	*/
int OutBytes,OutPos;		   /* Positioning indices for OutBuffer	*/

FILE *InPtr,*OutPtr;		      /* Input and output File Pointers	*/

char *InBuffer,*OutBuffer;		     /* File processing buffers	*/

char TargetPath[MAXPATH];	      /* Edited path of files to change	*/

char CurDir[MAXPATH];			   /* Current working directory	*/
char cdrive[MAXDRIVE],cdir[MAXDIR],cfile[MAXFILE],cext[MAXEXT];

DFILE *FileNameList;	   /* Base pointer for file name structure list	*/
DFILE *LastFileEntry;	   /* Pointer to file name structure last entry	*/

/************************************************************************/
/* External String and Function Declarations				*/
/************************************************************************/

extern char *sys_errlist[];	  		  /* DOS Error Messages	*/

/************************************************************************/
/* Global Utility Function Declarations.               			*/
/************************************************************************/

void SetCursor(int row,int col);
unsigned Row(),Col();

/************************************************************************/
/* FCHGALL Main Program Function						*/
/************************************************************************/

main(int argc, char *argv[])
{
	int Instructions(),Initialize(),MainLine();

	int rc;

	printf("\n%s\n",ProgTitle);		     /* Tell who we are	*/

	switch (argc)		 /* Capture the command line parameters	*/
		{
		case 3:
			strcpy(OldString,argv[1]);
			strcpy(NewString,argv[2]);
			break;
		case 4:
			strcpy(TargetPath,argv[1]);
			strcpy(OldString,argv[2]);
			strcpy(NewString,argv[3]);
			break;
		default:
			Instructions();		  /* Tell it like it is	*/
			exit (1);		 /* Set the return code	*/
		}
	
	if ( (rc = Initialization()) == SUCCESS )
		rc = MainLine();

	printf("\n\nFiles Read=%d  Files Changed=%d\n",FileCnt,ChgCnt);

	if ( (rc == FAIL) )
		abort();

	exit(0);
}

/************************************************************************/
/* Program Initialization routine					*/
/************************************************************************/

int Initialization()
{
	int EditPath(char *path),EditString(char *string);
	int LoadFileNames(struct ffblk *ffblk);

	int attr,lth;
	struct ffblk ffblk;	    /* Structure for findfirst/findnext	*/

	getcwd(CurDir,MAXPATH);	       /* Capture the current directory	*/

	lth = strlen(CurDir) - 1;	 /* Point to the last character	*/

	if ( (*(CurDir + lth) != BKSLASH)  )     /* Check for backslash	*/
		strcat(CurDir,"\\");		  /* Append a backslash	*/

	strcat(CurDir,"*.*");			/* Default to all files	*/

	fnsplit(CurDir,cdrive,cdir,cfile,cext);  /* Make CWD components	*/

	if ( (TargetPath[0] != NULL) )	/* Was there a designated path?	*/
		{
		if ( (EditPath(TargetPath)) == FAIL )    /* Validate it	*/
			return (FAIL);
		}
	else
		strcpy(TargetPath,CurDir);

	OldLength = EditString(OldString);
	NewLength = EditString(NewString);

/* Show exactly what we're changing after the editing has been done	*/

	printf("\nChanging /%s/ to /%s/\n",OldString,NewString);

	printf("\n%s%s",cdrive,cdir);		  /* Show the directory	*/

	row = Row();
	col = Col();		       /* Retain the file name position	*/

/* Allocate space for the input and output (double) buffers	*/

	if ( (InBuffer = (char *) calloc(2,RECSIZE)) == NULL )
		{
		printf("\nInsufficient Memory\n");
		return (FAIL);
		}

	if ( (OutBuffer = (char *) calloc(2,RECSIZE)) == NULL )
		{
		printf("\nInsufficient Memory\n");
		return (FAIL);
		}

/* Use the findfirst/findnext functions to load the structure list	*/

	attr = FA_DIREC | FA_SYSTEM | FA_HIDDEN | FA_RDONLY;

	if ( (findfirst(TargetPath,&ffblk,attr)) != 0 )
		{
		printf("\n%s\n%s\n",TargetPath,sys_errlist[errno]);
		return (FAIL);
		}

	do	{
		if ( (LoadFileNames(&ffblk)) == FAIL )
			return (FAIL);

		} while ( (findnext(&ffblk)) == 0 );

	return (SUCCESS);
}

int LoadFileNames(struct ffblk *ffblk)
{
	DFILE *fnm;

/* Make sure the file we're processing is a legitimate file	*/

	if ( (ffblk->ff_attrib & FA_DIREC) )
		return (SUCCESS);

	if ( (ffblk->ff_attrib & FA_SYSTEM) )
		return (SUCCESS);

	if ( (ffblk->ff_attrib & FA_HIDDEN) )
		return (SUCCESS);

	if ( (ffblk->ff_attrib & FA_RDONLY) )
		return (SUCCESS);

	FileCnt++;		      /* Count the file being processed	*/

	if ( (fnm = (DFILE *) calloc(1,sizeof(DFILE))) == NULL )
		{
		printf("\nInsufficient Memory for File Names\n");
		return (FAIL);
		}

	strncpy(fnm->filename,ffblk->ff_name,12); /* Load the file name	*/

	if ( (LastFileEntry != NULL) )
		LastFileEntry->next = fnm;     /* Link it into the list	*/

	if ( (FileNameList == NULL) )	/* If the base is NULL, this is	*/
		FileNameList = fnm;		     /* the first entry	*/

	LastFileEntry = fnm;			 /* Keep the last entry	*/

	return (SUCCESS);
}

int MainLine()
{
	int ProcessFileChanges(char *fname);

	DFILE *fnm;

/* Pass down through the structure list and process each one	*/

	for (fnm = FileNameList; fnm != NULL; fnm = fnm->next)
		if ( (ProcessFileChanges(fnm->filename)) == FAIL )
			return (FAIL);

	return (SUCCESS);
}

int ProcessFileChanges(char *fname)
{
	int ProcessRecordChanges();

	char inpath[MAXPATH];

	ChangesMade = FALSE;		       /* Reset the change flag	*/

	InBytes = InPos = OutBytes = OutPos = 0;   /* Clear the indices	*/

	SetCursor(row,col);
	printf("%-12s",fname);			  /* Show the file name	*/

	sprintf(inpath,"%s%s%s",cdrive,cdir,fname);   /* Build the path	*/

/* Open the input file for processing	*/

	if ( (InPtr = fopen(inpath,"rb")) == NULL )
		{
		printf("\n\n%s\n%s\n",inpath,sys_errlist[errno]);
		return (FAIL);
		}

/* Use RECSIZE for the item count because we need the exact bytes read	*/

	if ( (InBytes = fread(InBuffer,1,RECSIZE,InPtr)) == 0 ) /* Primed */
		{
		fclose(InPtr);
		return (SUCCESS);			  /* Empty File	*/
		}

/* Open a temporary file for the output records	*/

	if ( (OutPtr = fopen("temp.$$$","wb")) == NULL )
		{
		printf("\n\ntemp.$$$%s\n",sys_errlist[errno]);
		return (FAIL);
		}

	do	{
		ByteCnt = InBytes;		    /* Size of Buffer 1	*/

/* Load the second input buffer for the next time around	*/

		InBytes = fread(InBuffer+RECSIZE,1,RECSIZE,InPtr);

		ProcessRecordChanges();		    /* Work on Buffer 1	*/

		} while ( InBytes > 0 );   /* When Zero, we're all done	*/

	if ( (OutPos > 0) )	      /* Write out the last short block	*/
		if ( (fwrite(OutBuffer,1,OutPos,OutPtr)) < OutPos )
			{
			printf("\n\ntemp.$$$\n%s\n",sys_errlist[errno]);
			return (FAIL);		    /* Stop right there	*/
			}

	fclose(InPtr);				    /* Close both files	*/
	fclose(OutPtr);

	if ( (ChangesMade) )			/* The file has changed	*/
		{
		ChgCnt++;			 /* Count it as changed	*/

		unlink(inpath);			  /* Ditch the old file	*/

		if ( (rename("temp.$$$",inpath)) != 0 )  /* Rename TEMP	*/
			{
			printf("\n\ntemp.$$$\n%s\n",sys_errlist[errno]);
			return (FAIL);	   /* Leave everything as it is	*/
			}
		}
	else
		unlink("TEMP.$$$");	      /* Clean up the directory	*/

	return (SUCCESS);
}

int ProcessRecordChanges()
{
	int WriteOutput();
	void MoveBuffer(int curpos),ChangeString();

	int inch = InPos;

	while (inch < ByteCnt)		       /* Scan the first buffer	*/
		{
		if ( (memicmp(InBuffer+inch,OldString,OldLength)) == 0 )
			{
			MoveBuffer(inch);       /* Move input to output	*/
			ChangeString();		     /* Make the change	*/

			if ( (OutPos > RECSIZE) )  /* Check buffer full	*/
				if ( (WriteOutput()) == FAIL )
					return (FAIL);

			inch = InPos;	 /* Reposition after old string	*/
			}
		else
			inch++;		/* Advance to the next position	*/
		}	

/* Process the remainder of the input buffer without changes	*/

	if ( (inch > InPos) )
		MoveBuffer(inch);		/* Move input to output	*/

	if ( (OutPos > RECSIZE) )	     /* Drain the output buffer	*/
		if ( (WriteOutput()) == FAIL )
			return (FAIL);

	if ( (InBytes > 0) )	  	  /* Shift Buffer 2 to Buffer 1	*/
		{
		memcpy(InBuffer,InBuffer+RECSIZE,InBytes);
		memset(InBuffer+InBytes,NULL,RECSIZE);
		}

	InPos -= ByteCnt;

	return (SUCCESS);
}

void MoveBuffer(curpos)
{
	int movcnt;

	movcnt = curpos - InPos;	     /* Calculate bytes to move	*/

	memcpy(OutBuffer+OutPos,InBuffer+InPos,movcnt);	     /* Move it	*/

	InPos += movcnt;			  /* Adjust the indices	*/
	OutPos += movcnt;
}

void ChangeString()
{
	memcpy(OutBuffer+OutPos,NewString,NewLength);

	InPos += OldLength;			  /* Adjust the indices	*/
	OutPos += NewLength;

	ChangesMade = TRUE;
}

int WriteOutput()
{
	if ( (fwrite(OutBuffer,1,RECSIZE,OutPtr)) < RECSIZE )
		{
		printf("\n\ntemp.$$$\n%s\n",sys_errlist[errno]);
		return (FAIL);			    /* Stop right there	*/
		}

	OutPos -= RECSIZE;	      /* Reset the primary output index	*/

	memcpy(OutBuffer,OutBuffer+RECSIZE,RECSIZE);   /* Shift buffers	*/
	memset(OutBuffer+OutPos,NULL,RECSIZE);

	return (SUCCESS);
}

/************************************************************************/
/* User Instructions							*/
/************************************************************************/

int Instructions()
{
	printf("\nCommand Syntax: FCHGALL [filespec] oldstring newstring\n");

	printf("\nFCHGALL will scan multiple files in a directory changing any occurrences");
	printf("\nof oldstring to newstring.  If either string contains imbedded blanks,");
	printf("\nit must be surrounded with double quotes.  The scan is case insensitive.\n");

	printf("\nThe filespec is optional and may include the DOS wildcard characters.");
	printf("\nThe default is to all files in the current working directory.\n");

	printf("\n*************** WARNING  WARNING  WARNING *******************");
	printf("\nFCHGALL provides no confirmation option when changing strings");
	printf("\n*************** WARNING  WARNING  WARNING *******************\n");

	return (SUCCESS);
}

int EditPath(char *path)
{
	int attr,lth;
	char *c;
	char drive[MAXDRIVE],dir[MAXDIR],file[MAXFILE],ext[MAXEXT];
	struct ffblk ffblk;

	strupr(path);

/* Make sure the path supplied gets by DOS's validations	*/

	attr = FA_DIREC | FA_SYSTEM | FA_HIDDEN | FA_RDONLY;

	if ( (findfirst(path,&ffblk,attr)) != 0 )  /* Validate the path	*/
		{
		printf("\n%s\n%s\n",path,sys_errlist[errno]);
		return (FAIL);
		}

/* If the user supplied just a directory name, make sure the path ends	*/
/* with a valid file component.	*/

	if ( (ffblk.ff_attrib & FA_DIREC) == FA_DIREC )
		{
		if ( (strpbrk(path,"?*")) == NULL )   /* Any wildcards?	*/
			{
			lth = strlen(path) - 1;		/* Yes	*/

			if ( (*(path + lth) != BKSLASH) )
				strcat(path,"\\");

			strcat(path,"*.*");	   /* Dummy a file spec	*/
			}
		}

	fnsplit(path,drive,dir,file,ext);   /* Break it into components	*/

	if ( (dir[0] == '.') ) /* Check for a partial directory segment	*/
		{
		if ( (c = strrchr(cdir,BKSLASH)) != NULL )
			{
			*(c) = NULL;	    /* Kill cdir's trailing '\'	*/

			if ( (c = strrchr(cdir,BKSLASH)) != NULL )
				*(c) = NULL; /* and it's last component	*/

			strcpy(cdir,&dir[2]);	 /* Add partial dir seg	*/
			dir[0] = NULL;	    /* Kill the partial dir seg	*/
			}
		}

	if ( (strlen(drive)) == 0 )	 /* Did we have a drive letter?	*/
		strcpy(drive,cdrive);		 /* No, use the default	*/

	if ( (strlen(dir)) == 0 )    /* Did we have a directory string?	*/
		strcpy(dir,cdir);		 /* No, use the default	*/

	fnmerge(path,drive,dir,file,ext);        /* Put it all together	*/
	fnsplit(path,cdrive,cdir,cfile,cext);    /* Overlay the globals	*/

	return (SUCCESS);
}

int EditString(char *string)
{
	int lth = strlen(string);	       /* We'll need the length	*/

	if ( (*(string) != QUOTE) )		     /* Any Delimiter ?	*/
		return (lth);		  /* No, just return the length	*/

	if ( (*(string+lth-1) == *(string)) ) /* Are delimiters paired?	*/
		{
		*(string+lth-1) = NULL;       /* Yes, kill the last one	*/
		memcpy(string,string+1,lth-1);  /* Shift over the first	*/
		lth -= 2;			   /* Adjust the length	*/
		}

	return (lth);
}

/************************************************************************/
/* Global Utility Functions 						*/
/************************************************************************/

void SetCursor(int row,int col)
{
	_DL = col;
	_DH = row;
	_AH = 2;			 /* Function code to set cursor	*/
	_BH = 0;				/* Current display page	*/

	geninterrupt(0x10);
}

unsigned Row()				/* Gets the current display row	*/
{
	_AH = 3;			      /* Function to get cursor	*/
	_BH = 0;			        /* Current display page	*/

	geninterrupt(0x10);

	return (_DH);		       /* The row is in the dh register	*/
}

unsigned Col()			     /* Gets the current display column	*/
{
	_AH = 3;			      /* Function to get cursor	*/
	_BH = 0;			        /* Current display page	*/

	geninterrupt(0x10);

	return (_DL);		    /* The column is in the dl register	*/
}
